/*****************************************************************************
* chdlc_rec.c	CHDLC API: Receive Module
*
* Author(s):	Gideon Hack & Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 1995-1999 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <net/bpf.h>
#include <net/if.h>
#include <net/wanpipe.h>
#include <net/sdla_chdlc.h>
#include <netinet/in.h>

#define FALSE	0
#define TRUE	1

#define LGTH_CRC_BYTES	2
unsigned short no_CRC_bytes_Rx;
unsigned char HDLC_streaming = FALSE;


#define FRM_LGTH 50
#define HDR_LGTH (sizeof(api_rx_hdr_t) + sizeof(struct bpf_hdr) + 4)

unsigned char Rx_data[PRI_MAX_NO_DATA_BYTES_IN_FRAME + 
		      LGTH_CRC_BYTES + sizeof(api_rx_hdr_t)];
 

/* Prototypes */
void open_bpf(void);
int MakeConnection(char *);
void recv_socket( void);

int 			desc;
fd_set 			ready;
struct sockaddr_in 	sal;
struct sockaddr_in 	sad;
CMDBLOCK_STRUCT		cmdblock;
int 			counter=0;
pid_t			pid_number;


/*
 * Find and open bpf interface.
 */
void open_bpf(void)
{
    char device[sizeof"/dev/bpf000"];
    int n = 0;

    /* Go through all the minors and find one that isn't in use. */
    do {
	(void)sprintf(device, "/dev/bpf%d", n++);
	desc = open(device, O_RDONLY);
    } while(desc < 0 && errno == EBUSY && n < 4);
    
    if (desc < 0)
    {
	fprintf(stderr, "Error: All BPF device are busy\n");
	exit(0);
    }
    return;
}

/***************************************************
* MakeConnection
*
*   o Create a Socket
*   o Bind a socket to a wanpipe network interface
*       (Interface name is supplied by the user)
*/         

int MakeConnection( char *i_name ) 
{
	struct ifreq ifr;
  	u_int imm_mode = 1;
	u_int len;

        no_CRC_bytes_Rx = (HDLC_streaming) ? LGTH_CRC_BYTES : 0;
	len = FRM_LGTH + no_CRC_bytes_Rx + HDR_LGTH; 
	if (ioctl(desc, BIOCSBLEN, &len) < 0)
	{
	    fprintf(stderr, "Error: Cannot set buffer length %d\n", len);
	    exit(0);
	}
	
	if (ioctl(desc, BIOCIMMEDIATE, &imm_mode) < 0)
	{
	    fprintf(stderr, "Error: Cannot set immediate mode\n");
	    exit(0);
	}
	
	strncpy(ifr.ifr_name, i_name, sizeof(ifr.ifr_name));
	if (ioctl(desc, BIOCSETIF, &ifr) < 0)
	{
	    fprintf(stderr, "Error: Unknown interface name %s (Err-%d)\n",
						i_name, errno);
	    exit(0);
	}

	printf("Socket bound to %s\n",i_name);
	return( TRUE );

}

/***************************************************
* RECV_SOCKET 
*
*   o Read a socket 
*   o Cast data received to api_rx_element_t data type 
*   o The received packet contains 16 bytes header 
*
*	------------------------------------------
*      |  16 bytes      |        X bytes        ...
*	------------------------------------------
* 	   Header              Data
*
*   o Data structures:
*
*	typedef struct {
*        	unsigned char   error_flag      PACKED;
*        	unsigned short  time_stamp      PACKED;
*        	unsigned char   reserved[13]    PACKED;
*	} api_rx_hdr_t; 
*
*	typedef struct {
*        	api_rx_hdr_t    api_rx_hdr      PACKED;
*        	void *          data            PACKED;
*	} api_rx_element_t;
*
*
*/

void recv_socket(void) 
{
    unsigned long Rx_count;
    unsigned short Rx_lgth;
    api_rx_element_t* api_rx_el;
    u_int len = 0, i = 0;
    int err;

    Rx_count = 0;

    /* If running HDLC_STREAMING then the received CRC bytes
     * will be passed to the application as part of the 
     * received data.  
     */
    no_CRC_bytes_Rx = (HDLC_streaming) ? LGTH_CRC_BYTES : 0;
    len = FRM_LGTH + no_CRC_bytes_Rx + HDR_LGTH;

    for(;;) 
    {	
	FD_ZERO(&ready);
	FD_SET(desc,&ready);

  	if (select(desc + 1,&ready, NULL, NULL, NULL))
	{
	    err = read(desc, Rx_data, len);
	    /*err = 
	     *	read(desc, Rx_data, 
	     *	     sizeof(Rx_data) + sizeof(api_rx_hdr_t) + 
	     *	     sizeof(struct bpf_hdr));
	     */

	    /* err indicates bytes received */
	    if (err != -1) 
	    {
		/* Check the packet length */
		Rx_lgth = err - HDR_LGTH;
		if (Rx_lgth <= no_CRC_bytes_Rx) 
		{
		    printf("\nShort frame received (%d)\n", Rx_lgth);
		    return;
		}

		Rx_lgth -= no_CRC_bytes_Rx;

		/* Cast data received from the socket to a
                 * api_rx_element_t data type 
		 *       api_rx_el->api_rx_hdr.error_flag : 
		 *			Error Flags  (1 bytes)
 		 *       api_rx_el->api_rx_hdr.time_stamp : 
		 * 			Time Stamp   (2 bytes)
		 *			in miliseconds and wraps
       		 *                      over 65535 ms.
		 *       api_rx_el->data : 
		 *                      Points to data portion of the packet 
		 */
		api_rx_el = (api_rx_element_t *)&Rx_data[sizeof(struct bpf_hdr)+4];


		/* Print out packet information and number of packets received */
		printf("\nReceived Packet #%lu, len: %d, err: 0x%02X, tm: %5u\n", 
				++Rx_count, Rx_lgth, 
				api_rx_el->api_rx_hdr.error_flag, 
				api_rx_el->api_rx_hdr.time_stamp);

		/* Print out the receive data */
                /*for (i = 0;i < Rx_lgth; i ++) 
                 *   printf("%x ", *((unsigned char *)&api_rx_el->data + i));
                 *printf("\n");
		 */ 

	    } 
	    else 
	    {
		printf("\nError receiving data (%d)\n", errno);
		break;
	    }
	} 
	else 
	{
            printf("\nError selecting socket\n");
	    break;
	}
    }
} 



/***************************************************************
 * Main:
 *
 *    o Make a socket connection to the driver.
 *    o Call recv_socket() to read the socket 
 *
 **************************************************************/


int main(int argc, char* argv[])
{
	int proceed;

	if (argc != 2){
		printf("Usage: trec <interface name>\n");
		exit(0);
	}	

	open_bpf();
	proceed = MakeConnection(argv[argc-1]);
        pid_number = getpid();
	if( proceed == TRUE ){
		recv_socket( );
		return 0;
	}
	close(desc);
	return 1;
};
